import numpy as np


class Evaluator(object):
    def __init__(self, num_class):
        self.num_class = num_class
        self.confusion_matrix = np.zeros((self.num_class,)*2)

    def Pixel_Accuracy(self):
        Acc = np.diag(self.confusion_matrix).sum() / self.confusion_matrix.sum()
        return Acc

    def Pixel_Accuracy_Class(self):
        Acc = np.diag(self.confusion_matrix) / self.confusion_matrix.sum(axis=1)
        Acc = np.nanmean(Acc)
        return Acc

    def Mean_Intersection_over_Union(self):
        MIoU = np.diag(self.confusion_matrix) / (
                    np.sum(self.confusion_matrix, axis=1) + np.sum(self.confusion_matrix, axis=0) -
                    np.diag(self.confusion_matrix))
        MIoU = np.nanmean(MIoU)
        return MIoU

    def Frequency_Weighted_Intersection_over_Union(self):
        freq = np.sum(self.confusion_matrix, axis=1) / np.sum(self.confusion_matrix)
        iu = np.diag(self.confusion_matrix) / (
                    np.sum(self.confusion_matrix, axis=1) + np.sum(self.confusion_matrix, axis=0) -
                    np.diag(self.confusion_matrix))

        FWIoU = (freq[freq > 0] * iu[freq > 0]).sum()
        return FWIoU

    def _generate_matrix(self, gt_image, pre_image):
        
        mask = (gt_image >= 0) & (gt_image < self.num_class)
        
        label = self.num_class * gt_image[mask].astype('int') + pre_image[mask]
        #print(gt_image[mask].shape)
        count = np.bincount(label, minlength=self.num_class**2)
        confusion_matrix = count.reshape(self.num_class, self.num_class)
        
        #print(MIoU)
        #print(confusion_matrix)
        return confusion_matrix

    def add_batch(self, gt_image, pre_image):
        #print(gt_image.shape,pre_image.shape)

        assert gt_image.shape == pre_image.shape
        self.confusion_matrix += self._generate_matrix(gt_image, pre_image)

    def reset(self):
        self.confusion_matrix = np.zeros((self.num_class,) * 2)







from src.ecolaf import *
import torch
import torch.nn as nn
#import torchvision
from segformer import Segformer
from functools import partial
#from dataloader import make_data_loader
from tqdm import tqdm
import numpy as np
#from JeanZayTraining.utils.metrics import Evaluator

NUM_CLASSES = 25 #19 #25
CUDA = True

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")


model_rgb = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1,
    ).to(device)
        

model_depth = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1, 
    in_chans=3,
    ).to(device)

model_lidar = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1,
    in_chans=3,
    ).to(device)

model_event = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1,
    in_chans=3,
    ).to(device)

        

ckpt = torch.load('checkpoint_DELIVER.pth.tar', weights_only=False)

rgb = {}
depth = {}
lidar = {}
event = {}

for k, v in ckpt["state_dict"].items():
    if 'model_rgb' in k:
        rgb[k[10:]] = v
    elif 'model_depth' in k:
        depth[k[12:]] = v
    elif 'model_lidar' in k:
        lidar[k[12:]] = v
    elif 'model_event' in k:
        event[k[12:]] = v
    else:
        print(k)

model_rgb.load_state_dict(rgb)
model_depth.load_state_dict(depth)
model_lidar.load_state_dict(lidar)
model_event.load_state_dict(event)

print('LOADING OK')

MyModel = ECOLAF([model_rgb, model_depth, model_lidar, model_event], NUM_CLASSES).cuda()


torch.save({'state_dict': MyModel.state_dict()}, 'new_ECOLAF_DELIVER_checkpoint.pth.tar')
print('checkpoint saved')


"""
MyModel = torch.load('checkpoint.pth.tar', weights_only = False).cuda()
for m in MyModel.models:
    m = m.cuda()
"""

ckpt = torch.load('new_ECOLAF_DELIVER_checkpoint.pth.tar')["state_dict"]
MyModel.load_state_dict(ckpt)

print('LOADING accomplished, sir!')


class Path(object):
    @staticmethod
    def db_root_dir(dataset):
        if dataset == 'DELIVER':
            return '/home/lderegnaucourt/data/DELIVER/'
        elif dataset == 'MUSES' or dataset == 'MUSES_CLEAR':
            return 'home/lderegnaucourt/data/MUSES/'
        else:
            print('Dataset {} not available.'.format(dataset))
            raise NotImplementedError



import os
import shutil
import torch
from collections import OrderedDict
import glob

class Saver(object):


    def __init__(self, args):
        self.args = args
        self.directory = os.path.join('run', args.dataset, args.checkname)
        self.runs = sorted(sorted(glob.glob(os.path.join(self.directory, 'experiment_*'))), key=len)
        run_id = int(len(self.runs))+1 if self.runs else 0
        print("run_id: ", run_id)
        print("runs: ", self.runs)

        self.experiment_dir = os.path.join(self.directory, 'experiment_{}'.format(str(run_id)))
        if not os.path.exists(self.experiment_dir):
            os.makedirs(self.experiment_dir)

    def save_checkpoint(self, state, is_best, filename='checkpoint.pth.tar'):
        """Saves checkpoint to disk"""
        filename = os.path.join(self.experiment_dir, filename)
        torch.save(state, filename)
        if is_best:
            best_pred = state['best_pred']
            with open(os.path.join(self.experiment_dir, 'best_pred.txt'), 'w') as f:
                f.write(str(best_pred))
            if self.runs:
                previous_miou = [0.0]
                for run in self.runs:
                    run_id = run.split('_')[-1]
                    path = os.path.join(self.directory, 'experiment_{}'.format(str(run_id)), 'best_pred.txt')
                    if os.path.exists(path):
                        with open(path, 'r') as f:
                            #print(f.readline())
                            miou = float(f.readline().strip())
                            #miou = f.readline()
                            previous_miou.append(miou)
                    else:
                        continue
                max_miou = max(previous_miou)
                if best_pred > max_miou:
                    shutil.copyfile(filename, os.path.join(self.directory, 'model_best.pth.tar'))
            else:
                shutil.copyfile(filename, os.path.join(self.directory, 'model_best.pth.tar'))

    def save_experiment_config(self):
        logfile = os.path.join(self.experiment_dir, 'parameters.txt')
        log_file = open(logfile, 'w')
        p = OrderedDict()
        p['dataset'] = self.args.dataset
        p['backbone'] = self.args.backbone
        p['out_stride'] = self.args.out_stride
        p['lr'] = self.args.lr
        p['lr_scheduler'] = self.args.lr_scheduler
        p['loss_type'] = self.args.loss_type
        p['epoch'] = self.args.epochs
        p['batch_size'] = self.args.batch_size
        p['use_pretrained_resnet'] = self.args.use_pretrained_resnet
        p['base_size'] = self.args.base_size
        p['crop_size'] = self.args.crop_size

        for key, val in p.items():
            log_file.write(key + ':' + str(val) + '\n')
        log_file.close()



##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
## Created by: Hang Zhang
## ECE Department, Rutgers University
## Email: zhang.hang@rutgers.edu
## Copyright (c) 2017
##
## This source code is licensed under the MIT-style license found in the
## LICENSE file in the root directory of this source tree
##+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import math
from torch.optim.lr_scheduler import _LRScheduler
import torch

class WarmupLR(_LRScheduler):
    def __init__(self, optimizer, warmup_iter=500, warmup_ratio=5e-4, warmup='exp', last_epoch=-1) -> None:
        self.warmup_iter = warmup_iter
        self.warmup_ratio = warmup_ratio
        self.warmup = warmup
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        ratio = self.get_lr_ratio()
        return [ratio * lr for lr in self.base_lrs]

    def get_lr_ratio(self):
        return self.get_warmup_ratio() if self.last_epoch < self.warmup_iter else self.get_main_ratio()

    def get_main_ratio(self):
        raise NotImplementedError

    def get_warmup_ratio(self):
        assert self.warmup in ['linear', 'exp']
        alpha = self.last_epoch / self.warmup_iter

        return self.warmup_ratio + (1. - self.warmup_ratio) * alpha if self.warmup == 'linear' else self.warmup_ratio ** (1. - alpha)


class WarmupPolyLR(WarmupLR):
    def __init__(self, optimizer, power, max_iter, warmup_iter=500, warmup_ratio=5e-4, warmup='exp', last_epoch=-1) -> None:
        self.power = power
        self.max_iter = max_iter
        super().__init__(optimizer, warmup_iter, warmup_ratio, warmup, last_epoch)

    def get_main_ratio(self):
        real_iter = self.last_epoch - self.warmup_iter
        real_max_iter = self.max_iter - self.warmup_iter
        alpha = real_iter / real_max_iter

        return ((1 - alpha) ** self.power).real

class WarmupCosLR(WarmupLR):
    def __init__(self, optimizer, max_iter, warmup_iter=500, warmup_ratio=5e-4, warmup='exp', last_epoch=-1) -> None:
        self.max_iter = max_iter
        super().__init__(optimizer, warmup_iter, warmup_ratio, warmup, last_epoch)

    def get_main_ratio(self):
        real_iter = self.last_epoch - self.warmup_iter
        real_max_iter = self.max_iter - self.warmup_iter
        alpha = real_iter / real_max_iter

        return 0.5 * (1 + math.cos(1.0 *alpha * math.pi)) 


class LR_Scheduler(object):
    """Learning Rate Scheduler

    Step mode: ``lr = baselr * 0.1 ^ {floor(epoch-1 / lr_step)}``

    Cosine mode: ``lr = baselr * 0.5 * (1 + cos(iter/maxiter))``

    Poly mode: ``lr = baselr * (1 - iter/maxiter) ^ 0.9``

    Args:
        args:
          :attr:`args.lr_scheduler` lr scheduler mode (`cos`, `poly`),
          :attr:`args.lr` base learning rate, :attr:`args.epochs` number of epochs,
          :attr:`args.lr_step`

        iters_per_epoch: number of iterations per epoch
    """
    def __init__(self, mode, base_lr, num_epochs, iters_per_epoch=0,
                 lr_step=0, warmup_epochs=0):
        self.mode = mode
        print('Using {} LR Scheduler!'.format(self.mode))
        self.lr = base_lr
        if mode == 'step':
            assert lr_step
        self.lr_step = lr_step
        self.iters_per_epoch = iters_per_epoch
        self.N = num_epochs * iters_per_epoch
        self.epoch = -1
        self.warmup_iters = warmup_epochs * iters_per_epoch

    def __call__(self, optimizer, i, epoch, best_pred):
        T = epoch * self.iters_per_epoch + i
        if self.mode == 'cos':
            lr = 0.5 * self.lr * (1 + math.cos(1.0 * T / self.N * math.pi))
        elif self.mode == 'poly':
            lr = self.lr * pow((1 - 1.0 * T / self.N), 0.9)
        elif self.mode == 'step':
            lr = self.lr * (0.5 ** (epoch // self.lr_step))
        else:
            raise NotImplemented
        # warm up lr schedule
        if self.warmup_iters > 0 and T < self.warmup_iters:
            lr = lr * 1.0 * T / self.warmup_iters
        if epoch > self.epoch:
            print('\n=>Epoches %i, learning rate = %.4f, \
                previous best = %.4f' % (epoch, lr, best_pred))
            self.epoch = epoch
        assert lr >= 0
        self._adjust_learning_rate(optimizer, lr)

    def _adjust_learning_rate(self, optimizer, lr):
        if len(optimizer.param_groups) == 1:
            optimizer.param_groups[0]['lr'] = lr
        else:
            # enlarge the lr at the head
            optimizer.param_groups[0]['lr'] = lr
            for i in range(1, len(optimizer.param_groups)):
                optimizer.param_groups[i]['lr'] = lr * 10



from src.ecolaf import *
import torch
import torch.nn as nn
#import torchvision
from segformer import Segformer
from functools import partial
from models.dataloaders import make_data_loader
from tqdm import tqdm
import numpy as np
from models.utils.metrics import Evaluator
import random, os

NUM_CLASSES = 25
CUDA = True

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
torch.autograd.set_detect_anomaly(True)

torch.set_float32_matmul_precision(precision="medium")

model_rgb = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1,
    ).to(device)
        
model_depth = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1, 
    in_chans=3,
    ).to(device)

model_lidar = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1,
    in_chans=3,
    ).to(device)

model_event = Segformer(
    pretrained = None,
    img_size = 512, 
    patch_size = 4, 
    embed_dims = (64, 128, 320, 512), 
    num_heads = (1, 2, 5, 8), 
    mlp_ratios = (4, 4, 4, 4),
    qkv_bias = True, 
    depths = (3, 4, 6, 3), 
    sr_ratios = (8, 4, 2, 1),
    drop_rate = 0.0, 
    drop_path_rate = 0.1,
    decoder_dim = 512,
    norm_layer = partial(nn.LayerNorm, eps=1e-6), 
    num_classes=NUM_CLASSES+1,
    in_chans=3,
    ).to(device)


device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

MyModel = ECOLAF([model_rgb, model_depth, model_lidar, model_event], NUM_CLASSES).to(device)

ckpt = torch.load('ECOLAF_DELIVER_checkpoint.pth.tar')

MyModel.load_state_dict(ckpt['state_dict'], strict=True)

train_loader, val_loader, test_loader = make_data_loader(dataset='DELIVER', root='/home/lderegnaucourt/data/DELIVER', crop_size=800, batch_size=1, test_batch_size=2)

tbar = tqdm(test_loader, desc='\r')

MyModel.eval()

evaluator = Evaluator(NUM_CLASSES)
evaluator.reset()

for i, (sample, label) in enumerate(tbar):
    image, target, depth, lidar, event = sample[0].to(device), label.to(device), sample[1].to(device), sample[2].to(device), sample[3].to(device)
    with torch.amp.autocast('cuda', enabled=False):
        with torch.no_grad():
            output = MyModel([image, depth, lidar, event], interpolation=True)
    pred = output.data.cpu().numpy()
    target_ = target.cpu().numpy()
    pred = np.argmax(pred, axis=1)
    evaluator.add_batch(target_, pred)

mIoU = evaluator.Mean_Intersection_over_Union()
print("mIoU:{}".format(mIoU))
confusion_matrix = evaluator.confusion_matrix
np.save('confusion_matrix',confusion_matrix)



__author__='Lucas DEREGNAUCOURT'
__author_email__='lucas.deregnaucourt@insa-rouen.fr'

import torch

class Dempster_layer(torch.nn.Module):
    '''
    Apply the scalable Dempster fusion
    '''
    def __init__(self):
        super(Dempster_layer, self).__init__()
   
    def forward(self,m):
        m_omega = torch.cat((m[:,:-1,:,:,:]+m[:,None,-1,:,:,:],m[:,None,-1,:,:,:]),1)
        tmp = torch.sum(torch.log(torch.relu(m_omega)+1e-15),2)
        tmp = tmp-tmp.max(1,keepdim=True).values
        
        m1 = torch.exp(tmp)
        prodOmega = m1[:,-1,None]
        m1 = m1[...,:-1, :, :] - prodOmega
        
        m1 = torch.cat((m1,prodOmega),1)
        m1 = torch.nn.functional.normalize(m1,p=1.,dim=1)
        return m1


class DSmP(torch.nn.Module):
    '''
    Apply the Dezert-Smarandache probability transformation
    '''
    def __init__(self, eps):
        super(DSmP, self).__init__()
        self.eps = eps
   
    def forward(self, x):
        return (x+(x*x[:,None,-1, :, :]+self.eps*x[:,None,-1, :, :])/(1-x[:,None,-1, :, :]+self.eps*(x.shape[1]-1)))[:,:-1]
     

class Discounting_layer(torch.nn.Module):
    '''
    Apply the adaptive discounting procedure:
    1. Compute the pairwise distances
    2. Compute the pairwise conflicts
    3. Compute the discounting coefficients
    4. Discount the mass functions
    '''
    def __init__(self,nb_classe,nb_expert):
        super(Discounting_layer,self).__init__()

        self.nb_classe = nb_classe
        self.nb_expert = nb_expert

        '''
        Indices matrix for the pairwise distances computation
        '''
        count = 0
        indices = torch.zeros((nb_expert, nb_expert-1), dtype=int)
        for i in range(nb_expert-1):
            for j in range(i+1, nb_expert):
                    indices[i, j-1] = count
                    indices[j, i] = count
                    count+=1

        tab_indices = torch.triu_indices(self.nb_expert,self.nb_expert,offset=1)
        self.const = 1 - (2*self.nb_classe+1)/(self.nb_classe+1)**2 #Constant used to compute the conflicts from the distances
        self.lamb = 2 #lambda parameter for the discounting coefficients computation

        self.register_buffer("tab_indices", tab_indices)
        self.register_buffer("indices", indices)

    def forward(self,x, output_conflict=False, output_discouting_coefs=False):      

        '''
        Jousselme's distance
        '''
        m = torch.index_select(x,2,self.tab_indices[0]) - torch.index_select(x,2,self.tab_indices[1])
        m = torch.sum(m**2,1) + 2./self.nb_classe * m[:,-1,...]*torch.sum(m[:,:-1,...],1)

        '''
        Conflicts derived from the distances
        '''
        conf = self.const * torch.sqrt(m+1e-15/2.)

        '''
        Discouting coefficients derived from the conflicts
        '''
        conf_per_modality = conf[:, self.indices, :, :].sum(dim=2) + 1e-15
        conf_per_modality = (conf_per_modality / (conf_per_modality.shape[1]-1))
        discountings = ((1-conf_per_modality.unsqueeze(1)**self.lamb)**(1/self.lamb))

        '''
        Mass functions discounting
        '''
        new_x_om = x[:, :-1, :, :, :] * discountings
        new_x = torch.clamp(torch.cat((new_x_om, 1-new_x_om.sum(dim=1).unsqueeze(dim=1)), dim=1), min=0., max=1.)

        return new_x, conf_per_modality if output_conflict else None, discountings.squeeze(1) if output_discouting_coefs else None




import math
import torch
import torch.nn as nn
import torch.nn.functional as F

from timm.models.layers import DropPath, to_2tuple, trunc_normal_
#from models.segformer_utils.logger import get_root_logger
#from mmcv.runner import load_checkpoint



class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = self.fc1(x)
        x = self.dwconv(x, H, W)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0., sr_ratio=1):
        super().__init__()
        assert dim % num_heads == 0, f"dim {dim} should be divided by num_heads {num_heads}."

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        B, N, C = x.shape
        q = self.q(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
            x_ = self.sr(x_).reshape(B, C, -1).permute(0, 2, 1)
            x_ = self.norm(x_)
            kv = self.kv(x_).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        else:
            kv = self.kv(x).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, sr_ratio=1):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale,
            attn_drop=attn_drop, proj_drop=drop, sr_ratio=sr_ratio)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class OverlapPatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """

    def __init__(self, img_size=224, patch_size=7, stride=4, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)

        self.img_size = img_size
        self.patch_size = patch_size
        self.H, self.W = img_size[0] // patch_size[0], img_size[1] // patch_size[1]
        self.num_patches = self.H * self.W
        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=stride,
                              padding=(patch_size[0] // 2, patch_size[1] // 2))
        self.norm = nn.LayerNorm(embed_dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x):
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return x, H, W
    
    
class LinearMLP(nn.Module):
    """
    Linear Embedding
    """
    def __init__(self, input_dim=2048, embed_dim=768):
        super().__init__()
        self.proj = nn.Linear(input_dim, embed_dim)

    def forward(self, x):
        x = x.flatten(2).transpose(1, 2)
        x = self.proj(x)
        return x
    
    

class DWConv(nn.Module):
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        return x
    
    
class Segformer(nn.Module):
    def __init__(
        self, 
        pretrained=None,
        img_size=1024, 
        patch_size=4, 
        in_chans=3, 
        num_classes=19, 
        embed_dims=[64, 128, 320, 512],
        num_heads=[1, 2, 5, 8], 
        mlp_ratios=[4, 4, 4, 4], 
        qkv_bias=True, 
        qk_scale=None, 
        drop_rate=0.,
        attn_drop_rate=0., 
        drop_path_rate=0., 
        norm_layer=nn.LayerNorm,
        depths=[3, 6, 40, 3], 
        sr_ratios=[8, 4, 2, 1],
        decoder_dim = 768
    ):
        super().__init__()
        self.num_classes = num_classes
        self.depths = depths
        self.backbone = nn.Module()
        # patch_embed
        self.backbone.patch_embed1 = OverlapPatchEmbed(
            img_size=img_size, 
            patch_size=7, 
            stride=4, 
            in_chans=in_chans,
            embed_dim=embed_dims[0]
        )
        self.backbone.patch_embed2 = OverlapPatchEmbed(
            img_size=img_size // 4, 
            patch_size=3, 
            stride=2, 
            in_chans=embed_dims[0],
            embed_dim=embed_dims[1]
        )
        self.backbone.patch_embed3 = OverlapPatchEmbed(
            img_size=img_size // 8, 
            patch_size=3, 
            stride=2, 
            in_chans=embed_dims[1],
            embed_dim=embed_dims[2]
        )
        self.backbone.patch_embed4 = OverlapPatchEmbed(
            img_size=img_size // 16, 
            patch_size=3, 
            stride=2, 
            in_chans=embed_dims[2],
            embed_dim=embed_dims[3]
        )

        # transformer encoder
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule
        cur = 0
        self.backbone.block1 = nn.ModuleList([Block(
            dim=embed_dims[0], num_heads=num_heads[0], mlp_ratio=mlp_ratios[0], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer, sr_ratio=sr_ratios[0])
            for i in range(depths[0])])
        self.backbone.norm1 = norm_layer(embed_dims[0])

        cur += depths[0]
        self.backbone.block2 = nn.ModuleList([Block(
            dim=embed_dims[1], num_heads=num_heads[1], mlp_ratio=mlp_ratios[1], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer, sr_ratio=sr_ratios[1])
            for i in range(depths[1])])
        self.backbone.norm2 = norm_layer(embed_dims[1])

        cur += depths[1]
        self.backbone.block3 = nn.ModuleList([Block(
            dim=embed_dims[2], num_heads=num_heads[2], mlp_ratio=mlp_ratios[2], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer, sr_ratio=sr_ratios[2])
            for i in range(depths[2])])
        self.backbone.norm3 = norm_layer(embed_dims[2])

        cur += depths[2]
        self.backbone.block4 = nn.ModuleList([Block(
            dim=embed_dims[3], num_heads=num_heads[3], mlp_ratio=mlp_ratios[3], qkv_bias=qkv_bias, qk_scale=qk_scale,
            drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[cur + i], norm_layer=norm_layer, sr_ratio=sr_ratios[3])
            for i in range(depths[3])])
        self.backbone.norm4 = norm_layer(embed_dims[3])

        # classification head
        # self.head = nn.Linear(embed_dims[3], num_classes) if num_classes > 0 else nn.Identity()
        
        
        # segmentation head
        self.linear_c4 = LinearMLP(input_dim=embed_dims[3], embed_dim=decoder_dim)
        self.linear_c3 = LinearMLP(input_dim=embed_dims[2], embed_dim=decoder_dim)
        self.linear_c2 = LinearMLP(input_dim=embed_dims[1], embed_dim=decoder_dim)
        self.linear_c1 = LinearMLP(input_dim=embed_dims[0], embed_dim=decoder_dim)
        self.linear_fuse = nn.Conv2d(4 * decoder_dim, decoder_dim, 1, bias=False)
        self.linear_fuse_bn = nn.BatchNorm2d(decoder_dim)
        self.activate = nn.ReLU(True)
        self.dropout = nn.Dropout2d(drop_rate)
        self.linear_pred = nn.Conv2d(decoder_dim, num_classes, kernel_size=1)

        self.apply(self._init_weights)
        self.init_weights(pretrained=pretrained)
        

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def init_weights(self, pretrained=None):
        pass
        """
        if isinstance(pretrained, str):
            logger = get_root_logger()
            load_checkpoint(self, pretrained, map_location='cpu', strict=False, logger=logger)
        """
    def reset_drop_path(self, drop_path_rate):
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(self.depths))]
        cur = 0
        for i in range(self.depths[0]):
            self.block1[i].drop_path.drop_prob = dpr[cur + i]

        cur += self.depths[0]
        for i in range(self.depths[1]):
            self.block2[i].drop_path.drop_prob = dpr[cur + i]

        cur += self.depths[1]
        for i in range(self.depths[2]):
            self.block3[i].drop_path.drop_prob = dpr[cur + i]

        cur += self.depths[2]
        for i in range(self.depths[3]):
            self.block4[i].drop_path.drop_prob = dpr[cur + i]

    def freeze_patch_emb(self):
        self.patch_embed1.requires_grad = False

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed1', 'pos_embed2', 'pos_embed3', 'pos_embed4', 'cls_token'}  # has pos_embed may be better

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=''):
        self.num_classes = num_classes
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

    def forward_features(self, x):
        B = x.shape[0]
        outs = []

        # stage 1
        x, H, W = self.backbone.patch_embed1(x)
        for i, blk in enumerate(self.backbone.block1):
            x = blk(x, H, W)
        x = self.backbone.norm1(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 2
        x, H, W = self.backbone.patch_embed2(x)
        for i, blk in enumerate(self.backbone.block2):
            x = blk(x, H, W)
        x = self.backbone.norm2(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 3
        x, H, W = self.backbone.patch_embed3(x)
        for i, blk in enumerate(self.backbone.block3):
            x = blk(x, H, W)
        x = self.backbone.norm3(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 4
        x, H, W = self.backbone.patch_embed4(x)
        for i, blk in enumerate(self.backbone.block4):
            x = blk(x, H, W)
        x = self.backbone.norm4(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        return outs

    def forward(self, x):
        x = self.forward_features(x)
        
        c1, c2, c3, c4 = x 

         ############## MLP decoder on C1-C4 ###########
        n, _, h, w = c4.shape
        h_out, w_out = c1.size()[2], c1.size()[3]

        _c4 = self.linear_c4(c4).permute(0,2,1).reshape(n, -1, c4.shape[2], c4.shape[3])
        _c4 = F.interpolate(_c4, size = c1.size()[2:], mode = 'bilinear', align_corners = False)

        _c3 = self.linear_c3(c3).permute(0,2,1).reshape(n, -1, c3.shape[2], c3.shape[3])
        _c3 = F.interpolate(_c3, size = c1.size()[2:], mode = 'bilinear', align_corners = False)

        _c2 = self.linear_c2(c2).permute(0,2,1).reshape(n, -1, c2.shape[2], c2.shape[3])
        _c2 = F.interpolate(_c2, size = c1.size()[2:], mode = 'bilinear', align_corners = False)

        _c1 = self.linear_c1(c1).permute(0,2,1).reshape(n, -1, c1.shape[2], c1.shape[3])

        _c = self.linear_fuse(torch.cat([_c4, _c3, _c2, _c1], dim = 1))
        _c = self.linear_fuse_bn(_c)
        _c = self.activate(_c)

        x = self.dropout(_c)
        x = self.linear_pred(x)
        
        x = F.interpolate(input = x, size = (h_out, w_out), mode = 'bilinear', align_corners = False)
        x = x.type(torch.float32)
    
        return x



import torch
import torch.nn as nn
import numpy as np
from torch import Tensor

KITTI_LOSS_WEIGHT = torch.tensor([
    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # background
    [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # asphalt
    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # concrete
    [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # metal
    [0.0, 0.5, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # road marking
    [0.0, 0.5, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # tar
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # fabric, leather
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # glass
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # plaster
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.4, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # plastic
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # rubber, vinyl
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # mud sand soil gravel
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # ceramic
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # cobblestone
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], # brick
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.1, 0.5, 0.0, 0.0, 0.0], # grass
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0], # wood
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.2, 1.0, 0.0, 0.0, 0.0], # leaf
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0], # water
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0], # human body
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0], # sky
],dtype=torch.float32)

class MyCrossEntropyLoss(torch.nn.NLLLoss):
    def __init__(self, *args,**kwargs) -> None:
        super().__init__(*args,**kwargs)

    def forward(self, input: Tensor, target: Tensor, mask: Tensor) -> Tensor:
        if mask is not None:
            assert target.size() == mask.size()
            target[mask>0] = self.ignore_index
        return super().forward(torch.log(input+1e-20),target)


class MyWeightedCrossEntropyLoss(nn.NLLLoss):
    def __init__(self, weight=None, size_average=None, ignore_index=-100, reduce=None, reduction='mean'):
        super(MyWeightedCrossEntropyLoss, self).__init__(weight, size_average, ignore_index, reduce, reduction)

    def forward(self, input, target, weight):
        """
        Args:
            input: Log-probabilities from your model (shape: [batch_size, num_classes, ...]).
            target: Ground truth labels (shape: [batch_size, ...]).
            weight: A tensor of weights (shape: [batch_size, ...]) for each sample.
        """
        if weight is None:
            raise ValueError("Weight tensor must be provided for weighted loss computation.")

        # Apply log softmax to input if it's not already in log-probability format
        if not torch.is_tensor(weight):
            weight = torch.tensor(weight, dtype=torch.float32, device=input.device)

        nll_loss = F.nll_loss(torch.log(input+1e-20), target, weight=self.weight, reduction='none')

        # Multiply by the sample weights
        weighted_loss = nll_loss * weight

        # Aggregate according to the specified reduction
        if self.reduction == 'mean':
            return weighted_loss.mean()
        elif self.reduction == 'sum':
            return weighted_loss.sum()
        else:
            return weighted_loss


class MyMSE(torch.nn.MSELoss):
    def __init__(self, *args,**kwargs) -> None:
        super().__init__(*args,**kwargs)

    def forward(self, input: Tensor, target: Tensor, mask: Tensor) -> Tensor:
        return super().forward(input,target)


class MyOhemCrossEntropy(nn.Module):
    def __init__(self, ignore_label: int = 255, weight: Tensor = None, thresh: float = 0.7, aux_weights: list = [1, 1]) -> None:
        super().__init__()
        self.ignore_label = ignore_label
        self.aux_weights = aux_weights
        self.thresh = -torch.log(torch.tensor(thresh, dtype=torch.float))
        self.criterion = nn.NLLLoss(weight=weight, ignore_index=ignore_label, reduction='none')

    def _forward(self, preds: Tensor, labels: Tensor, weights: Tensor, mask=None) -> Tensor:
        # preds in shape [B, C, H, W] and labels in shape [B, H, W]
        n_min = labels[labels != self.ignore_label].numel() // 16
        loss = (weights * self.criterion(preds, labels)).view(-1)
        loss_hard = loss[loss > self.thresh]

        if loss_hard.numel() < n_min:
            loss_hard, _ = loss.topk(n_min)

        return torch.mean(loss_hard)

    def forward(self, preds, labels: Tensor, weights: Tensor) -> Tensor:
        if isinstance(preds, tuple):
            return sum([w * self._forward(pred, labels, weights=weights) for (pred, w) in zip(preds, self.aux_weights)])
        return self._forward(preds, labels, weights=weights)

class SegmentationLosses(object):
    def __init__(self, weight=None, size_average=True, batch_average=False, ignore_index=255, cuda=False):
        self.ignore_index = ignore_index
        self.weight = weight
        self.size_average = size_average
        self.batch_average = batch_average
        self.cuda = cuda

    def build_loss(self, mode='ce'):
        """Choices: ['ce' or 'focal' or 'original']"""
        if mode == 'ce':
            # return self.CrossEntropyLoss
            return self.CrossEntropyLossAdv
        elif mode == 'focal':
            return self.FocalLoss
        elif mode == 'original':
            return self.OriginalLoss
        elif mode == 'bce':
            return self.BCELoss
        elif mode == 'my_ce':
            return self.MyCrossEntropyAdv
        elif mode == 'my_ohem':
            return self.MyOhemCrossEntropy
        else:
            raise NotImplementedError

    def CrossEntropyLoss(self, logit, target):
        n, c, h, w = logit.size()
        criterion = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index,
                                        size_average=self.size_average)
        if self.cuda:
            criterion = criterion.cuda()

        loss = criterion(torch.log(logit+1e-20), target.long())

        if self.batch_average:
            loss = loss/n

        return loss

    def MyCrossEntropyAdv(self, logit, target, mask=None):
        n, c, h, w = logit.size()
        if mask is not None:
            assert target.size() == mask.size()
            target[mask>0] = self.ignore_index
        criterion = nn.NLLLoss(weight=self.weight, ignore_index=self.ignore_index, size_average=self.size_average)
        if self.cuda:
            criterion = criterion.cuda()
        #loss=criterion(torch.log(logit+1e-20), target.long())
        loss = criterion(torch.log(torch.clamp(logit, 1e-20, 1)), target.long())
        if self.batch_average:
            loss = loss/n

        return loss


    def MyOhemCrossEntropy(self, logit, target, weights=None, mask=None, thresh: float = 0.7, aux_weights: list = [1, 1]):
            thresh = -torch.log(torch.tensor(thresh, dtype=torch.float))
            criterion = nn.NLLLoss(weight=self.weight, ignore_index=self.ignore_index, reduction='none')
            if self.cuda:
                criterion = criterion.cuda()


            def _forward(preds: Tensor, labels: Tensor, criterion, ignore_index=self.ignore_index) -> Tensor:
                # preds in shape [B, C, H, W] and labels in shape [B, H, W]
                n_min = labels[labels != ignore_index].numel() // 16
                if isinstance(weights, Tensor):
                    loss = (weights * criterion(torch.log(torch.clamp(preds, 1e-20, 1)), labels.long())).view(-1)
                else:
                    loss = criterion(torch.log(torch.clamp(preds, 1e-20, 1)), labels.long()).view(-1)
                loss_hard = loss[loss > thresh]

                if loss_hard.numel() < n_min:
                    loss_hard, _ = loss.topk(n_min)

                return torch.mean(loss_hard)

            if isinstance(logit, tuple):
                return sum([w * _forward(lo, labels, criterion) for (lo, w) in zip(logit, aux_weights)])
            return _forward(logit, target, criterion)



    def CrossEntropyLossAdv(self, logit, target, mask=None):
        n, c, h, w = logit.size()
        #print(target.long().size())
        #print(mask.size())
        if mask is not None:
            assert target.size() == mask.size()
            target[mask>0] = self.ignore_index
            #target[mask>0] = 0
        #print(self.ignore_index)
        criterion = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index,
                                        size_average=self.size_average)
        #criterion = nn.BCEWithLogitsLoss(weight=self.weight,
        #                                 size_average=self.size_average)
        if self.cuda:
            criterion = criterion.cuda()
        #print(logit.size())   
        #target=target.unsqueeze(1) 
        #loss = criterion(logit, target.long())
        #a=nn.Sigmoid()
        #print(a(logit))
        #print(target.long().size())
        
        loss = criterion(logit, target.long())
        #print(loss)
        if self.batch_average:
            loss = loss/n

        return loss
    def BCELoss(self, logit, target1, mask=None):
        n, c, h, w = logit.size()
        loss=0
        #print(target.long().size())
        #print(mask.size())
        if mask is not None:
            assert target1.size() == mask.size()
            target1[mask>0] = self.ignore_index
            #target[mask>0] = 0
        criterion = nn.BCEWithLogitsLoss(weight=self.weight,
                                         size_average=self.size_average)
        if self.cuda:
            criterion = criterion.cuda()
        target2=np.zeros((target1.shape[0],20,target1.shape[1],target1.shape[2]))
        for k in range(target1.shape[0]):
            for i in range(20):
                target2[k,i,:,:]=np.int64(target1.cpu().numpy()[k,:,:]==i)
        target=torch.from_numpy(target2).cuda()
        for i in range(20):
            #print(target[:,i,:,:].shape)
            loss += criterion(logit[:,i,:,:], target[:,i,:,:])
        #print(loss)
        if self.batch_average:
            loss = loss/n

        return loss

    def FocalLoss(self, logit, target, mask=None, gamma=2, alpha=0.5):
        n, c, h, w = logit.size()

        if mask is not None:
            assert target.size() == mask.size()
            target[mask>0] = self.ignore_index

        criterion = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index,
                                        size_average=self.size_average)
        if self.cuda:
            criterion = criterion.cuda()

        logpt = -criterion(logit, target.long())
        pt = torch.exp(logpt)
        if alpha is not None:
            logpt *= alpha
        loss = -((1 - pt) ** gamma) * logpt

        if self.batch_average:
            loss /= n

        return loss

    def OriginalLoss(self, logit, target):
        n, c, h, w = logit.size()
        criterion = nn.CrossEntropyLoss(weight=self.weight, ignore_index=self.ignore_index,
                                        size_average=self.size_average)
        if self.cuda:
            criterion = criterion.cuda()
        target_weight = KITTI_LOSS_WEIGHT[target.long()].permute((0,3,1,2)).float().cuda()
        exp_x  = torch.exp(logit)
        loss = torch.sum(torch.mean(-torch.sum(target_weight * logit,dim=1) + torch.log(torch.sum(exp_x,dim=1)),dim=(1,2)))

        if self.batch_average:
            loss /= n

        return loss

if __name__ == "__main__":
    loss = SegmentationLosses(cuda=True)
    a = torch.rand(1, 21, 7, 7).cuda()
    b = torch.rand(1, 7, 7).cuda()
    print(a)
    print(b)
    print(loss.CrossEntropyLoss(a, b).item())
    print(loss.FocalLoss(a, b, gamma=0, alpha=None).item())
    print(loss.OriginalLoss(a,b).item())
    print(loss.FocalLoss(a, b, gamma=2, alpha=0.5).item())







__author__='Lucas DEREGNAUCOURT'
__author_email__='lucas.deregnaucourt@insa-rouen.fr'

import torch
import torch.nn as nn
import torch.nn.functional as F

from Dempster_Shafer_utils import *

class ECOLAF(nn.Module):
    def __init__(self, models:list, num_classes:int):
        '''
        models: a list of M unimodal models.
        !!! Be sure that the linear layer output K+1 values (K: number of classes) !!!

        num_classes (Integer): number of classes
        '''
        super(ECOLAF, self).__init__()

        self.models = nn.ModuleList(models)
        self.fusion_layer = Dempster_layer()
        self.discounting_layer = Discounting_layer(num_classes, len(models))
        self.proba_transform = DSmP(1e-3)

    def forward(self, input:list, output_unimodal=False, output_conflict=False, output_discounting_coef=False, interpolation=False, output_keyword=None):
        '''
        We consider that M models were given for the initialization.
        
        input: list of M images
        
        output_unimodal (Boolean): returns the list of outputs for each model
        
        output_conflict (Boolean): returns a conflict map of shape BxMxHxW (B: batch size)
        
        output_discounting_coef (Boolean): returns the discounting coefficients map of shape BxMxHxW (B: batch size)
        
        interpolation (Boolean): interpolates the output to the original image size
        
        output_keyword (String): key to access to the output tensor if the model's output is a dictionary 
        (for example, the deeplabV3_resnet50 model from torchvision requires output_keyword='out')
        '''


        '''
        Infer the output for each modality
        '''
        unimodals = [] if output_unimodal else None        
        modality_outputs = []

        for i in range(len(self.models)):
            out_u = self.models[i](input[i])
            
            if output_keyword is not None:
                out_u = out_u[output_keyword]
            out_u = F.softmax(out_u, dim=1)
            if output_unimodal:
                unimodals.append(out_u)
            modality_outputs.append(out_u)

        out = torch.stack(modality_outputs, dim=2)
        
        '''
        Discount the mass functions with the adaptive discounting layer
        '''
        out_discounted, conflict_map, discounting_coefs_map = self.discounting_layer(out, output_conflict, output_discounting_coef)

        '''
        Merge the mass functions with the Dempster's rule
        '''
        out_fusion = self.fusion_layer(out_discounted)
            
        '''
        Resize the output to the original shape
        '''
        interp_size = input[0].shape[2:]  #shape for the interpolation

        if interpolation:
            out_fusion = F.interpolate(out_fusion, size=interp_size, mode='bilinear', align_corners=False)
        
        out_fusion = self.proba_transform(out_fusion)

        outputs = [out_fusion]

        if output_unimodal:
            if interpolation:
                unimodals = [self.proba_transform(F.interpolate(o, size=interp_size, mode='bilinear', align_corners=False)) for o in unimodals]
            else:
                unimodals = [self.proba_transform(o) for o in unimodals]
            outputs.append(unimodals)


        if output_conflict:
            outputs.append(conflict_map)
        if output_discounting_coef:
            outputs.append(discounting_coefs_map)
        
        return out_fusion if len(outputs)==1 else outputs



import torchvision.transforms.functional as TF 
import random
import math
import torch
from torch import Tensor
from typing import Tuple, List, Union, Tuple, Optional


class Compose:
    def __init__(self, transforms: list) -> None:
        self.transforms = transforms

    def __call__(self, sample: list) -> list:
        img, mask = sample['img'], sample['mask']
        if mask.ndim == 2:
            assert img.shape[1:] == mask.shape
        else:
            assert img.shape[1:] == mask.shape[1:]

        for transform in self.transforms:
            sample = transform(sample)

        return sample


class Normalize:
    def __init__(self, mean: list = (0.485, 0.456, 0.406), std: list = (0.229, 0.224, 0.225)):
        self.mean = mean
        self.std = std

    def __call__(self, sample: list) -> list:
        for k, v in sample.items():
            if k == 'mask':
                continue
            elif k == 'img':
                sample[k] = sample[k].float()
                sample[k] /= 255
                sample[k] = TF.normalize(sample[k], self.mean, self.std)
            else:
                if sample[k].dtype == torch.uint16:
                    sample[k] = sample[k].float()
                    sample[k] /= 65535
                else:
                    sample[k] = sample[k].float()
                    sample[k] /= 255
        
        return sample


class RandomColorJitter:
    def __init__(self, p=0.5) -> None:
        self.p = p

    def __call__(self, sample: list) -> list:
        if random.random() < self.p:
            self.brightness = random.uniform(0.5, 1.5)
            sample['img'] = TF.adjust_brightness(sample['img'], self.brightness)
            self.contrast = random.uniform(0.5, 1.5)
            sample['img'] = TF.adjust_contrast(sample['img'], self.contrast)
            self.saturation = random.uniform(0.5, 1.5)
            sample['img'] = TF.adjust_saturation(sample['img'], self.saturation)
        return sample


class AdjustGamma:
    def __init__(self, gamma: float, gain: float = 1) -> None:
        """
        Args:
            gamma: Non-negative real number. gamma larger than 1 make the shadows darker, while gamma smaller than 1 make dark regions lighter.
            gain: constant multiplier
        """
        self.gamma = gamma
        self.gain = gain

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        return TF.adjust_gamma(img, self.gamma, self.gain), mask


class RandomAdjustSharpness:
    def __init__(self, sharpness_factor: float, p: float = 0.5) -> None:
        self.sharpness = sharpness_factor
        self.p = p

    def __call__(self, sample: list) -> list:
        if random.random() < self.p:
            sample['img'] = TF.adjust_sharpness(sample['img'], self.sharpness)
        return sample


class RandomAutoContrast:
    def __init__(self, p: float = 0.5) -> None:
        self.p = p

    def __call__(self, sample: list) -> list:
        if random.random() < self.p:
            sample['img'] = TF.autocontrast(sample['img'])
        return sample


class RandomGaussianBlur:
    def __init__(self, kernel_size: int = 3, p: float = 0.5) -> None:
        self.kernel_size = kernel_size
        self.p = p

    def __call__(self, sample: list) -> list:
        if random.random() < self.p:
            sample['img'] = TF.gaussian_blur(sample['img'], self.kernel_size)
            # img = TF.gaussian_blur(img, self.kernel_size)
        return sample


class RandomHorizontalFlip:
    def __init__(self, p: float = 0.5) -> None:
        self.p = p

    def __call__(self, sample: list) -> list:
        if random.random() < self.p:
            for k, v in sample.items():
                sample[k] = TF.hflip(v)
            return sample
        return sample


class RandomVerticalFlip:
    def __init__(self, p: float = 0.5) -> None:
        self.p = p

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        if random.random() < self.p:
            return TF.vflip(img), TF.vflip(mask)
        return img, mask


class RandomGrayscale:
    def __init__(self, p: float = 0.5) -> None:
        self.p = p

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        if random.random() < self.p:
            img = TF.rgb_to_grayscale(img, 3)
        return img, mask


class Equalize:
    def __call__(self, image, label):
        return TF.equalize(image), label


class Posterize:
    def __init__(self, bits=2):
        self.bits = bits # 0-8
        
    def __call__(self, image, label):
        return TF.posterize(image, self.bits), label


class Affine:
    def __init__(self, angle=0, translate=[0, 0], scale=1.0, shear=[0, 0], seg_fill=0):
        self.angle = angle
        self.translate = translate
        self.scale = scale
        self.shear = shear
        self.seg_fill = seg_fill
        
    def __call__(self, img, label):
        return TF.affine(img, self.angle, self.translate, self.scale, self.shear, TF.InterpolationMode.BILINEAR, 0), TF.affine(label, self.angle, self.translate, self.scale, self.shear, TF.InterpolationMode.NEAREST, self.seg_fill) 


class RandomRotation:
    def __init__(self, degrees: float = 10.0, p: float = 0.2, seg_fill: int = 0, expand: bool = False) -> None:
        """Rotate the image by a random angle between -angle and angle with probability p

        Args:
            p: probability
            angle: rotation angle value in degrees, counter-clockwise.
            expand: Optional expansion flag. 
                    If true, expands the output image to make it large enough to hold the entire rotated image.
                    If false or omitted, make the output image the same size as the input image. 
                    Note that the expand flag assumes rotation around the center and no translation.
        """
        self.p = p
        self.angle = degrees
        self.expand = expand
        self.seg_fill = seg_fill

    def __call__(self, sample: list) -> list:
        random_angle = random.random() * 2 * self.angle - self.angle
        if random.random() < self.p:
            for k, v in sample.items():
                if k == 'mask':                
                    sample[k] = TF.rotate(v, random_angle, TF.InterpolationMode.NEAREST, self.expand, fill=self.seg_fill)
                else:
                    sample[k] = TF.rotate(v, random_angle, TF.InterpolationMode.BILINEAR, self.expand, fill=0)
            # img = TF.rotate(img, random_angle, TF.InterpolationMode.BILINEAR, self.expand, fill=0)
            # mask = TF.rotate(mask, random_angle, TF.InterpolationMode.NEAREST, self.expand, fill=self.seg_fill)
        return sample
    

class CenterCrop:
    def __init__(self, size: Union[int, List[int], Tuple[int]]) -> None:
        """Crops the image at the center

        Args:
            output_size: height and width of the crop box. If int, this size is used for both directions.
        """
        self.size = (size, size) if isinstance(size, int) else size

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        return TF.center_crop(img, self.size), TF.center_crop(mask, self.size)


class RandomCrop:
    def __init__(self, size: Union[int, List[int], Tuple[int]], p: float = 0.5) -> None:
        """Randomly Crops the image.

        Args:
            output_size: height and width of the crop box. If int, this size is used for both directions.
        """
        self.size = (size, size) if isinstance(size, int) else size
        self.p = p

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        H, W = img.shape[1:]
        tH, tW = self.size

        if random.random() < self.p:
            margin_h = max(H - tH, 0)
            margin_w = max(W - tW, 0)
            y1 = random.randint(0, margin_h+1)
            x1 = random.randint(0, margin_w+1)
            y2 = y1 + tH
            x2 = x1 + tW
            img = img[:, y1:y2, x1:x2]
            mask = mask[:, y1:y2, x1:x2]
        return img, mask


class Pad:
    def __init__(self, size: Union[List[int], Tuple[int], int], seg_fill: int = 0) -> None:
        """Pad the given image on all sides with the given "pad" value.
        Args:
            size: expected output image size (h, w)
            fill: Pixel fill value for constant fill. Default is 0. This value is only used when the padding mode is constant.
        """
        self.size = size
        self.seg_fill = seg_fill

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        padding = (0, 0, self.size[1]-img.shape[2], self.size[0]-img.shape[1])
        return TF.pad(img, padding), TF.pad(mask, padding, self.seg_fill)


class ResizePad:
    def __init__(self, size: Union[int, Tuple[int], List[int]], seg_fill: int = 0) -> None:
        """Resize the input image to the given size.
        Args:
            size: Desired output size. 
                If size is a sequence, the output size will be matched to this. 
                If size is an int, the smaller edge of the image will be matched to this number maintaining the aspect ratio.
        """
        self.size = size
        self.seg_fill = seg_fill

    def __call__(self, img: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor]:
        H, W = img.shape[1:]
        tH, tW = self.size

        # scale the image 
        scale_factor = min(tH/H, tW/W) if W > H else max(tH/H, tW/W)
        # nH, nW = int(H * scale_factor + 0.5), int(W * scale_factor + 0.5)
        nH, nW = round(H*scale_factor), round(W*scale_factor)
        img = TF.resize(img, (nH, nW), TF.InterpolationMode.BILINEAR, antialias=None)
        mask = TF.resize(mask, (nH, nW), TF.InterpolationMode.NEAREST, antialias=None)

        # pad the image
        padding = [0, 0, tW - nW, tH - nH]
        img = TF.pad(img, padding, fill=0)
        mask = TF.pad(mask, padding, fill=self.seg_fill)
        return img, mask 


class Resize:
    def __init__(self, size: Union[int, Tuple[int], List[int]]) -> None:
        """Resize the input image to the given size.
        Args:
            size: Desired output size. 
                If size is a sequence, the output size will be matched to this. 
                If size is an int, the smaller edge of the image will be matched to this number maintaining the aspect ratio.
        """
        self.size = size

    def __call__(self, sample:list) -> list:
        H, W = sample['img'].shape[1:]

        # scale the image 
        scale_factor = self.size[0] / min(H, W)
        nH, nW = round(H*scale_factor), round(W*scale_factor)
        for k, v in sample.items():
            if k == 'mask':                
                sample[k] = TF.resize(v, (nH, nW), TF.InterpolationMode.NEAREST, antialias=None)
            else:
                sample[k] = TF.resize(v, (nH, nW), TF.InterpolationMode.BILINEAR, antialias=None)
        # img = TF.resize(img, (nH, nW), TF.InterpolationMode.BILINEAR)
        # mask = TF.resize(mask, (nH, nW), TF.InterpolationMode.NEAREST)

        # make the image divisible by stride
        alignH, alignW = int(math.ceil(nH / 32)) * 32, int(math.ceil(nW / 32)) * 32
        
        for k, v in sample.items():
            if k == 'mask':                
                sample[k] = TF.resize(v, (alignH, alignW), TF.InterpolationMode.NEAREST, antialias=None)
            else:
                sample[k] = TF.resize(v, (alignH, alignW), TF.InterpolationMode.BILINEAR, antialias=None)
        # img = TF.resize(img, (alignH, alignW), TF.InterpolationMode.BILINEAR)
        # mask = TF.resize(mask, (alignH, alignW), TF.InterpolationMode.NEAREST)
        return sample


class RandomResizedCrop:
    def __init__(self, size: Union[int, Tuple[int], List[int]], scale: Tuple[float, float] = (0.5, 2.0), seg_fill: int = 0) -> None:
        """Resize the input image to the given size.
        """
        self.size = size
        self.scale = scale
        self.seg_fill = seg_fill

    def __call__(self, sample: list) -> list:
        # img, mask = sample['img'], sample['mask']
        H, W = sample['img'].shape[1:]
        tH, tW = self.size

        # get the scale
        ratio = random.random() * (self.scale[1] - self.scale[0]) + self.scale[0]
        # ratio = random.uniform(min(self.scale), max(self.scale))
        scale = int(tH*ratio), int(tW*4*ratio)
        # scale the image 
        scale_factor = min(max(scale)/max(H, W), min(scale)/min(H, W))
        nH, nW = int(H * scale_factor + 0.5), int(W * scale_factor + 0.5)
        # nH, nW = int(math.ceil(nH / 32)) * 32, int(math.ceil(nW / 32)) * 32
        for k, v in sample.items():
            if k == 'mask':                
                sample[k] = TF.resize(v, (nH, nW), TF.InterpolationMode.NEAREST, antialias=None)
            else:
                sample[k] = TF.resize(v, (nH, nW), TF.InterpolationMode.BILINEAR, antialias=None)

        # random crop
        margin_h = max(sample['img'].shape[1] - tH, 0)
        margin_w = max(sample['img'].shape[2] - tW, 0)
        y1 = random.randint(0, margin_h+1)
        x1 = random.randint(0, margin_w+1)
        y2 = y1 + tH
        x2 = x1 + tW
        for k, v in sample.items():
            sample[k] = v[:, y1:y2, x1:x2]

        # pad the image
        if sample['img'].shape[1:] != self.size:
            padding = [0, 0, tW - sample['img'].shape[2], tH - sample['img'].shape[1]]
            for k, v in sample.items():
                if k == 'mask':                
                    sample[k] = TF.pad(v, padding, fill=self.seg_fill)
                else:
                    sample[k] = TF.pad(v, padding, fill=0)

        return sample



def get_train_augmentation(size: Union[int, Tuple[int], List[int]], seg_fill: int = 0):
    return Compose([
        RandomColorJitter(p=0.2), # 
        RandomHorizontalFlip(p=0.5), #
        RandomGaussianBlur((3, 3), p=0.2), #
        RandomResizedCrop(size, scale=(0.5, 2.0), seg_fill=seg_fill), #
        Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))
    ])

def get_val_augmentation(size: Union[int, Tuple[int], List[int]]):
    return Compose([
        Resize(size),
        Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))
    ])


if __name__ == '__main__':
    h = 230
    w = 420
    sample = {}
    sample['img'] = torch.randn(3, h, w)
    sample['depth'] = torch.randn(3, h, w)
    sample['lidar'] = torch.randn(3, h, w)
    sample['event'] = torch.randn(3, h, w)
    sample['mask'] = torch.randn(1, h, w)
    aug = Compose([
        RandomHorizontalFlip(p=0.5),
        RandomResizedCrop((512, 512), antialias=None),
        Resize((224, 224), antialias=None),
        Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
    ])
    sample = aug(sample)
    for k, v in sample.items():
        print(k, v.shape)



from dataloaders.datasets import deliver, muses, muses_clear_train
from torch.utils.data import DataLoader
from dataloaders.augmentations_mm import *

def make_data_loader(dataset, root='PathToDELIVER', crop_size=1024, batch_size=2, test_batch_size=2, **kwargs):

    if dataset == 'DELIVER':
        modals=['img', 'depth', 'lidar', 'event']
        #cases=['cloud', 'fog', 'night', 'rain', 'sun', 'motionblur', 'overexposure', 'underexposure', 'lidarjitter', 'eventlowres']
        traintransform = get_train_augmentation((crop_size, crop_size), seg_fill=255)
        valtransform = get_val_augmentation((crop_size, crop_size))
        testtransform = get_val_augmentation((crop_size, crop_size)) #(1024,1024) original image shape
        train_set = deliver.DELIVER(transform=traintransform, root=root, split='train', modals=modals)
        val_set = deliver.DELIVER(transform=valtransform, root=root, split='val', modals=modals)
        test_set = deliver.DELIVER(transform=testtransform, root=root, split='test', modals=modals)

        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True, **kwargs)
        val_loader = DataLoader(val_set, batch_size=batch_size, shuffle=False, **kwargs)
        test_loader = DataLoader(test_set, batch_size=test_batch_size, shuffle=False, **kwargs)

        return train_loader, val_loader, test_loader


    elif dataset == 'MUSES':
        modals=['img', 'event', 'lidar', 'radar']
        traintransform = get_train_augmentation((700, 1250), seg_fill=255)
        valtransform = get_val_augmentation((700, 1250))
        testtransform = get_val_augmentation((700, 1250))  #(1080,1920) original shape
        train_set = muses.MUSES(transform=traintransform, split='train', modals=modals)
        val_set = muses.MUSES(transform=valtransform, split='val', modals=modals)
        test_set = muses.MUSES(transform=testtransform, split='test', modals=modals)

        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True, **kwargs)
        val_loader = DataLoader(val_set, batch_size=batch_size, shuffle=False, **kwargs)
        test_loader = DataLoader(test_set, batch_size=test_batch_size, shuffle=False, **kwargs)

        return train_loader, val_loader, test_loader

    elif dataset == 'MUSES_CLEAR_TRAIN':
        modals=['img', 'event', 'lidar', 'radar']
        traintransform = get_train_augmentation((700, 1250), seg_fill=255)
        valtransform = get_val_augmentation((700, 1250))
        testtransform = get_val_augmentation((700,1250)) #(1080,1920) original shape
        train_set = muses_clear_train.MUSES_CLEAR_TRAIN(transform=traintransform, split='train', modals=modals, case='clear')
        val_set = muses_clear_train.MUSES_CLEAR_TRAIN(transform=valtransform, split='val', modals=modals, case='clear')
        test_set = muses_clear_train.MUSES_CLEAR_TRAIN(transform=testtransform, split='test', modals=modals, case='rainy') #CHANGER LES CONDITIONS METEO ICI ['rainy', 'fog', 'snow']

        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True, **kwargs)
        val_loader = DataLoader(val_set, batch_size=batch_size, shuffle=False, **kwargs)
        test_loader = DataLoader(test_set, batch_size=test_batch_size, shuffle=False, **kwargs)

        return train_loader, val_loader, test_loader
    else:
        raise NotImplementedError



import matplotlib.pyplot as plt
import numpy as np
import torch

def decode_seg_map_sequence(label_masks, dataset='pascal'):
    rgb_masks = []
    for label_mask in label_masks:
        rgb_mask = decode_segmap(label_mask, dataset)
        rgb_masks.append(rgb_mask)
    rgb_masks = torch.from_numpy(np.array(rgb_masks).transpose([0, 3, 1, 2]))
    return rgb_masks


def decode_segmap(label_mask, dataset, plot=False):
    """Decode segmentation class labels into a color image
    Args:
        label_mask (np.ndarray): an (M,N) array of integer values denoting
          the class label at each spatial location.
        plot (bool, optional): whether to show the resulting color image
          in a figure.
    Returns:
        (np.ndarray, optional): the resulting decoded color image.
    """
    if dataset == 'pascal' or dataset == 'coco':
        n_classes = 21
        label_colours = get_pascal_labels()
    elif dataset == 'kitti' or dataset == 'kitti_advanced' or dataset == 'kitti_advanced_manta' \
            or dataset == 'handmade_dataset' or dataset == 'handmade_dataset_stereo' or dataset == 'multimodal_dataset':
        n_classes = 20
        label_colours = get_my_labels()
    elif dataset == 'cityscapes':
        n_classes = 19
        label_colours = get_cityscapes_labels()
    elif dataset == "DELIVER":
        n_classes = 25
        label_colours = get_DELIVER_labels()
    else:
        raise NotImplementedError

    r = label_mask.copy()
    g = label_mask.copy()
    b = label_mask.copy()
    for ll in range(0, n_classes):
        r[label_mask == ll] = label_colours[ll, 0]
        g[label_mask == ll] = label_colours[ll, 1]
        b[label_mask == ll] = label_colours[ll, 2]
    r[label_mask == 255] = 0
    g[label_mask == 255] = 0
    b[label_mask == 255] = 0
    rgb = np.zeros((label_mask.shape[0], label_mask.shape[1], 3))
    rgb[:, :, 0] = r / 255.0
    rgb[:, :, 1] = g / 255.0
    rgb[:, :, 2] = b / 255.0
    if plot:
        plt.imshow(rgb)
        plt.show()
    else:
        return rgb


def encode_segmap(mask):
    """Encode segmentation label images as pascal classes
    Args:
        mask (np.ndarray): raw segmentation label image of dimension
          (M, N, 3), in which the Pascal classes are encoded as colours.
    Returns:
        (np.ndarray): class map with dimensions (M,N), where the value at
        a given location is the integer denoting the class index.
    """
    mask = mask.astype(int)
    label_mask = np.zeros((mask.shape[0], mask.shape[1]), dtype=np.int16)
    for ii, label in enumerate(get_pascal_labels()):
        label_mask[np.where(np.all(mask == label, axis=-1))[:2]] = ii
    label_mask = label_mask.astype(int)
    return label_mask


def get_cityscapes_labels():
    return np.array([
        [128, 64, 128],
        [244, 35, 232],
        [70, 70, 70],
        [102, 102, 156],
        [190, 153, 153],
        [153, 153, 153],
        [250, 170, 30],
        [220, 220, 0],
        [107, 142, 35],
        [152, 251, 152],
        [0, 130, 180],
        [220, 20, 60],
        [255, 0, 0],
        [0, 0, 142],
        [0, 0, 70],
        [0, 60, 100],
        [0, 80, 100],
        [0, 0, 230],
        [119, 11, 32]])


def get_pascal_labels():
    """Load the mapping that associates pascal classes with label colors
    Returns:
        np.ndarray with dimensions (21, 3)
    """
    return np.asarray([[0, 0, 0], [128, 0, 0], [0, 128, 0], [128, 128, 0],
                       [0, 0, 128], [128, 0, 128], [0, 128, 128], [128, 128, 128],
                       [64, 0, 0], [192, 0, 0], [64, 128, 0], [192, 128, 0],
                       [64, 0, 128], [192, 0, 128], [64, 128, 128], [192, 128, 128],
                       [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0],
                       [0, 64, 128]])

def get_my_labels():
    " r,g,b"
    return np.array([
        [ 44, 160,  44], # asphalt
        [ 31, 119, 180], # concrete
        [255, 127,  14], # metal
        [214,  39,  40], # road marking
        [140,  86,  75], # fabric, leather
        [127, 127, 127], # glass
        [188, 189,  34], # plaster
        [255, 152, 150], # plastic
        [ 23, 190, 207], # rubber
        [174, 199, 232], # sand
        [196, 156, 148], # gravel
        [197, 176, 213], # ceramic
        [247, 182, 210], # cobblestone
        [199, 199, 199], # brick
        [219, 219, 141], # grass
        [158, 218, 229], # wood
        [ 57,  59, 121], # leaf
        [107, 110, 207], # water
        [156, 158, 222], # human body
        [ 99, 121,  57]]) # sky

def get_DELIVER_labels():
    return np.array([
            [70, 70, 70],
            [100, 40, 40],
            [55, 90, 80],
            [220, 20, 60],
            [153, 153, 153],
            [157, 234, 50],
            [128, 64, 128],
            [244, 35, 232],
            [107, 142, 35],
            [0, 0, 142],
            [102, 102, 156],
            [220, 220, 0],
            [70, 130, 180],
            [81, 0, 81],
            [150, 100, 100],
            [230, 150, 140],
            [180, 165, 180],
            [250, 170, 30],
            [110, 190, 160],
            [170, 120, 50],
            [45, 60, 150],
            [145, 170, 100],
            [  0,  0, 230],
            [  0, 60, 100],
            [  0,  0, 70],
            ])




if __name__ == "__main__":
    from src.ecolaf import *
    from torchvision import models
    import matplotlib.pyplot as plt
    import numpy as np

    m1 = models.segmentation.deeplabv3_resnet50(num_classes=21)
    m2 = models.segmentation.deeplabv3_resnet50(num_classes=21)
    m3 = models.segmentation.deeplabv3_resnet50(num_classes=21)
    m4 = models.segmentation.deeplabv3_resnet50(num_classes=21)

    img1 = torch.rand(4, 3, 500, 500)
    img2 = torch.rand(4, 3, 500, 500)
    img3 = torch.rand(4, 3, 500, 500)
    img4 = torch.rand(4, 3, 500, 500)

    m = ECOLAF([m1, m2, m3, m4], num_classes=20)

    out, out_u, conf_map, discounting_map = m([img1, img2, img3, img4], output_unimodal=True, output_conflict=True, output_discounting_coef=True, output_keyword='out')
    conf_map = conf_map.detach()
    discounting_map = discounting_map.detach()

    fig, axs = plt.subplots(2, 2, figsize=(10, 8))
    axs[0, 0].imshow(conf_map[0, 0, ...], cmap='viridis')
    axs[0, 0].set_title('conflict 1')

    axs[0, 1].imshow(conf_map[0, 1, ...], cmap='viridis')
    axs[0, 1].set_title('Conflict 2')

    axs[1, 0].imshow(conf_map[0, 2, ...], cmap='viridis')
    axs[1, 0].set_title('Conflict 3')

    axs[1, 1].imshow(conf_map[0, 3, ...], cmap='viridis')
    axs[1, 1].set_title('Conflict 4')

    plt.tight_layout()
    plt.show()

    fig, axs = plt.subplots(2, 2, figsize=(10, 8))
    axs[0, 0].imshow(discounting_map[0, 0, ...], cmap='viridis')
    axs[0, 0].set_title('Discounting 1')

    axs[0, 1].imshow(discounting_map[0, 1, ...], cmap='viridis')
    axs[0, 1].set_title('Discounting 2')

    axs[1, 0].imshow(discounting_map[0, 2, ...], cmap='viridis')
    axs[1, 0].set_title('Discounting 3')

    axs[1, 1].imshow(discounting_map[0, 3, ...], cmap='viridis')
    axs[1, 1].set_title('Discounting 4')

    plt.tight_layout()
    plt.show()

    pred = out[0].data.cpu().numpy()
    pred = np.argmax(pred, axis=0)

    pred1 = out_u[0][0].data.cpu().numpy()
    pred1 = np.argmax(pred1, axis=0)

    pred2 = out_u[1][0].data.cpu().numpy()
    pred2 = np.argmax(pred2, axis=0)

    pred3 = out_u[2][0].data.cpu().numpy()
    pred3 = np.argmax(pred3, axis=0)

    pred4 = out_u[3][0].data.cpu().numpy()
    pred4 = np.argmax(pred4, axis=0)

    fig, axs = plt.subplots(2, 2, figsize=(10, 8))
    axs[0, 0].imshow(pred1, cmap='viridis')
    axs[0, 0].set_title('Prediction 1')

    axs[0, 1].imshow(pred2, cmap='viridis')
    axs[0, 1].set_title('Prediction 2')

    axs[1, 0].imshow(pred3, cmap='viridis')
    axs[1, 0].set_title('Prediction 3')

    axs[1, 1].imshow(pred4, cmap='viridis')
    axs[1, 1].set_title('Prediction 4')

    plt.tight_layout()
    plt.show()


    plt.imshow(pred)
    plt.title('fusion prediction')
    plt.show()



from setuptools import setup

######################################################################################################
################ You May Remove All the Comments Once You Finish Modifying the Script ################
######################################################################################################

setup(

    name = 'ecolaf', 
    
    version = '1.0.0',
    
    description = 'A toolkit python package to build an adaptive multimodal late fusion pipeline based on Dempster-Shafer Theory for semantic segmentation and classification tasks.',
    
    py_modules = ["ecolaf", "Dempster_Shafer_utils"],

    package_dir = {'':'src'},
    
    author = 'Lucas Deregnaucourt',
    author_email = 'lucas.deregnaucourt@insa-rouen.fr',
    
    long_description = open('README.md').read() + '\n\n' + open('CHANGELOG.md').read(),
    long_description_content_type = "text/markdown",
    
    url='https://github.com/deregnaucourtlucas/ECOLAF',
    
    include_package_data=True,

    install_requires = [

        'torch'

    ],
    
    keywords = ['Dempster-Shafer Theory', 'Deep Learning', 'Multimodal', 'Semantic Segmentation', 'Classification'],
    
)



"""
# Code adapted from:
# https://github.com/facebookresearch/Detectron/blob/master/detectron/utils/collections.py

Source License
# Copyright (c) 2017-present, Facebook, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################
#
# Based on:
# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------
"""

class AttrDict(dict):

    IMMUTABLE = '__immutable__'

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__[AttrDict.IMMUTABLE] = False

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        elif name in self:
            return self[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if not self.__dict__[AttrDict.IMMUTABLE]:
            if name in self.__dict__:
                self.__dict__[name] = value
            else:
                self[name] = value
        else:
            raise AttributeError(
                'Attempted to set "{}" to "{}", but AttrDict is immutable'.
                format(name, value)
            )

    def immutable(self, is_immutable):
        """Set immutability to is_immutable and recursively apply the setting
        to all nested AttrDicts.
        """
        self.__dict__[AttrDict.IMMUTABLE] = is_immutable
        # Recursively set immutable state
        for v in self.__dict__.values():
            if isinstance(v, AttrDict):
                v.immutable(is_immutable)
        for v in self.values():
            if isinstance(v, AttrDict):
                v.immutable(is_immutable)

    def is_immutable(self):
        return self.__dict__[AttrDict.IMMUTABLE]



import os
import torch
from torchvision.utils import make_grid
from tensorboardX import SummaryWriter
from dataloaders.utils import decode_seg_map_sequence

class TensorboardSummary(object):
    def __init__(self, directory):
        print('-------------- save DIR -----------------------')
        print(directory)
        self.directory = directory

    def create_summary(self):
        writer = SummaryWriter(log_dir=os.path.join(self.directory))
        return writer

    def visualize_image(self, writer, dataset, image, target, output, global_step):
        grid_image = make_grid(image[:3].clone().cpu().data, 3, normalize=True)
        writer.add_image('Image', grid_image, global_step)
        grid_image = make_grid(decode_seg_map_sequence(torch.max(output[:3], 1)[1].detach().cpu().numpy(),
                                                       dataset=dataset), 3, normalize=False, value_range=(0, 255))
        writer.add_image('Predicted label', grid_image, global_step)
        grid_image = make_grid(decode_seg_map_sequence(torch.squeeze(target[:3], 1).detach().cpu().numpy(),
                                                       dataset=dataset), 3, normalize=False, value_range=(0, 255))
        writer.add_image('Groundtruth label', grid_image, global_step)

    def visualize_validation_image(self, writer, dataset, image, target, output, global_step):
        if global_step==0:
            grid_image = make_grid(image[:10].clone().cpu().data, 10, normalize=True)
            writer.add_image('Val Image', grid_image, global_step)
            grid_image = make_grid(decode_seg_map_sequence(torch.squeeze(target[:10], 1).detach().cpu().numpy(),
                                                           dataset=dataset), 10, normalize=False, value_range=(0, 255))
            writer.add_image('Val Groundtruth label', grid_image, global_step)
       
        grid_image = make_grid(decode_seg_map_sequence(torch.max(output[:10], 1)[1].detach().cpu().numpy(),
                                                       dataset=dataset), 10, normalize=False, value_range=(0, 255))
        writer.add_image('Val Predicted label', grid_image, global_step)

    def visualize_test_image(self, writer, dataset, image, target, output, global_step):
        if global_step==0:
            grid_image = make_grid(image.clone().cpu().data, image.size(0), normalize=True)
            writer.add_image('Test Image', grid_image, global_step)
        
            grid_image = make_grid(decode_seg_map_sequence(torch.squeeze(target, 1).detach().cpu().numpy(),
                                                           dataset=dataset), target.size(0), normalize=False, value_range=(0, 255))
            writer.add_image('Test Groundtruth label', grid_image, global_step)

        grid_image = make_grid(decode_seg_map_sequence(torch.max(output.float(), 1)[1].detach().cpu().numpy(),
                                                       dataset=dataset), output.size(0), normalize=False, value_range=(0, 255))
        writer.add_image('Test Predicted label', grid_image, global_step)



import os
from io import BytesIO
import torch 
import numpy as np
from torch import Tensor
from torch.utils.data import Dataset
import torchvision.transforms.functional as TF 
from torchvision import io
from pathlib import Path
from typing import Tuple
import glob
import einops
from torch.utils.data import DataLoader
from torch.utils.data import DistributedSampler, RandomSampler
from mypath import Path
#from semseg.augmentations_mm import get_train_augmentation

class MUSES_CLEAR_TRAIN(Dataset):
    """
    num_classes: 19
    """
    NUM_CLASSES = 19
    CLASSES = ['road', 'sidewalk', 'building', 'wall', 
    'fence', 'pole', 'traffic_light', 'traffic_sign', 'vegetation', 
    'terrain', 'sky', 'person', 'rider', 'car', 
    'truck', 'bus', 'train', 'motorcycle', 'bicycle']

    PALETTE = torch.tensor([[70, 70, 70],
            [100, 40, 40],
            [55, 90, 80],
            [220, 20, 60],
            [153, 153, 153],
            [157, 234, 50],
            [128, 64, 128],
            [244, 35, 232],
            [107, 142, 35],
            [0, 0, 142],
            [102, 102, 156],
            [220, 220, 0],
            [70, 130, 180],
            [81, 0, 81],
            [150, 100, 100],
            [230, 150, 140],
            [180, 165, 180],
            [250, 170, 30],
            [110, 190, 160],
            ])
    #Path.db_root_dir('SHIFT')
    def __init__(self, root: str = Path.db_root_dir('MUSES_CLEAR'), split: str = 'train', transform = None, modals = ['img'], case = None) -> None:
        super().__init__()
        assert split in ['train', 'val', 'test']
        self.transform = transform
        self.n_classes = len(self.CLASSES)
        self.ignore_label = 255
        self.modals = modals
        self.files = sorted(glob.glob(os.path.join(*[root, '*', '*', 'img', '*', '*'])))
        #img:jpg, depth and semseg:png, flow:npz
        # --- debug
        # self.files = sorted(glob.glob(os.path.join(*[root, 'img', '*', split, '*', '*.png'])))[:100]
        # --- split as case
        if case is not None:
            assert case in ['clear', 'fog', 'rainy', 'snow'], "Case name not available."
            _temp_files = [f for f in self.files if case in f]
            self.files = _temp_files
        if split == 'train':
            _temp_files = [f for f in self.files if ('train/' in f or 'val/' in f)]
        elif split == 'val':
            _temp_files = [f for f in self.files if 'test/' in f]
        #elif split == 'test':
        #    if case == 'clear':
        #        _temp_files = [f for f in self.files if 'test/' in f]
        self.files = _temp_files
        if not self.files:
            raise Exception(f"No images found in {root}")
        print(f"Found {len(self.files)} {split} {case} images.")

    def __len__(self) -> int:
        return len(self.files)
    
    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor]:
        rgb = str(self.files[index])
        x1 = rgb.replace('/img', '/event').replace('_frame_camera', '_event_camera')
        x2 = rgb.replace('/img', '/lidar').replace('_frame_camera', '_lidar')
        x3 = rgb.replace('/img', '/radar').replace('_frame_camera', '_radar')
        lbl_path = rgb.replace('/img', '/GT').replace('_frame_camera', '_gt_labelTrainIds')

        sample = {}
        sample['img'] = io.read_image(rgb)[:3, ...]
        H, W = sample['img'].shape[1:]
        if 'event' in self.modals:
            sample['event'] = self._open_img(x1)
        if 'lidar' in self.modals:
            sample['lidar'] = self._open_img(x2)
        if 'radar' in self.modals:
            sample['radar'] = self._open_img(x3)
        label = io.read_image(lbl_path)[0,...].unsqueeze(0)
        sample['mask'] = label
        
        if self.transform:
            sample = self.transform(sample)
        label = sample['mask']
        del sample['mask']
        label = self.encode(label.squeeze().numpy()).long()
        sample = [sample[k] for k in self.modals]
        return sample, label

    def _open_img(self, file):
        img = io.decode_image(file)
        C, H, W = img.shape
        if C == 4:
            img = img[:3, ...]
        if C == 1:
            img = img.repeat(3, 1, 1)
        return img

    def _open_npz(self, file):
        img = np.load(file)   #(BytesIO(file))
        #print("img shape: ", img.shape)
        img = torch.as_tensor(img["flow"], dtype=torch.float32).permute(2, 0, 1)
        #print("img shape after permute", img.shape)
        return img.repeat(2, 1, 1)[:3, ...]

    def encode(self, label: Tensor) -> Tensor:
        return torch.from_numpy(label)


if __name__ == '__main__':
    cases = ['clear']
    traintransform = None #get_train_augmentation((1024, 1024), seg_fill=255)
    for case in cases:

        trainset = MUSES_CLEAR_TRAIN(transform=traintransform, split='val', case='clear')
        trainloader = DataLoader(trainset, batch_size=2, num_workers=2, drop_last=False, pin_memory=False)

        for i, (sample, lbl) in enumerate(trainloader):
            print(torch.unique(lbl))



import os
from tqdm import tqdm
import numpy as np
from mypath import Path

def calculate_weigths_labels(dataset, dataloader, num_classes):
    # Create an instance from the data loader
    z = np.zeros((num_classes,))
    # Initialize tqdm
    tqdm_batch = tqdm(dataloader)
    print('Calculating classes weights')
    for sample in tqdm_batch:
        y = sample['label']
        y = y.detach().cpu().numpy()
        mask = (y >= 0) & (y < num_classes)
        labels = y[mask].astype(np.uint8)
        count_l = np.bincount(labels, minlength=num_classes)
        z += count_l
    tqdm_batch.close()
    total_frequency = np.sum(z)
    class_weights = []
    for frequency in z:
        class_weight = 1 / (np.log(1.02 + (frequency / total_frequency)))
        class_weights.append(class_weight)
    ret = np.array(class_weights)
    classes_weights_path = os.path.join(Path.db_root_dir(dataset), dataset+'_classes_weights.npy')
    classes_weights_path1 = os.path.join(Path.db_root_dir(dataset), dataset+'_classes_weights1.npy')
    np.save(classes_weights_path, ret)
    np.save(classes_weights_path1, z)

    return ret